@@ -1,4 +1,4 @@
-# $Id: DBI.pm,v 1.13 2000/05/24 06:58:27 schwern Exp $
+# $Id: DBI.pm,v 1.19 2000/09/08 18:51:45 schwern Exp $
 
 package Class::DBI;
 
@@ -7,10 +7,12 @@ require 5.00502;
 use strict;
 
 use vars qw($VERSION);
-$VERSION = '0.13';
+$VERSION = '0.18';
 
 use Carp::Assert;
-use base qw(Class::Accessor Class::Data::Inheritable Ima::DBI);
+use base qw(Class::Accessor Class::Data::Inheritable Ima::DBI 
+            Class::Fields::Fuxor Class::Fields);
+use Class::Fields::Attribs;
 
 use protected qw(__Changed __AutoCommit);
 
@@ -67,9 +69,10 @@ sub _safe_exists {
 
   # Tell Class::DBI a little about yourself.
   Film->table('Movies');
-  Film->columns('All', qw( Title Director Rating NumExplodingSheep ));
-  Film->columns('Primary', 'Title');
-  Film->set_db('Main', 'dbi:mysql', 'me', 'noneofyourgoddamnedbusiness');
+  Film->columns(All     => qw( Title Director Rating NumExplodingSheep ));
+  Film->columns(Primary => qw( Title ));
+  Film->set_db('Main', 'dbi:mysql', 'me', 'noneofyourgoddamnedbusiness',
+               {AutoCommit => 1});
 
 
   #-- Meanwhile, in a nearby piece of code! --#
@@ -132,7 +135,9 @@ persistant.  More details about individual methods will follow.
 
 You must have an existing database set up, have DBI.pm installed and
 the necessary DBD:: driver module for that database.  See L<DBI> and
-the documentation of your particular database for details.
+the documentation of your particular database for details.  
+
+DBD::CSV works in a pinch.
 
 =item I<Set up a table for your objects to be stored in.>
 
@@ -167,7 +172,7 @@ match the columns in your database, one to one.  Class::DBI (via
 Class::Accessor) will use this information to determine how to create
 accessors.
 
-  Film->columns('All', qw( Title Director Rating NumExplodingSheep ));
+  Film->columns(All => qw( Title Director Rating NumExplodingSheep ));
 
 For more information about how you can more efficiently declare your columns,
 L<"Lazy Population of Columns">
@@ -186,7 +191,7 @@ will be the primary key in your database.  Class::DBI needs
 this piece of information in order to construct the proper SQL
 statements to access your stored objects.
 
-  Film->columns('Primary', 'Title');
+  Film->columns(Primary => 'Title');
 
 =item I<Declare a database connection>
 
@@ -194,7 +199,7 @@ Class::DBI needs to know how to access the database.  It does
 this through a DBI connection which you set up.  Set up is by calling
 the set_db() method and declaring a database connection named 'Main'.
 
-  Film->set_db('Main', 'dbi:mysql', 'user', 'password');
+  Film->set_db('Main', 'dbi:mysql', 'user', 'password', {AutoCommit => 1});
 
 set_db() is inherited from Ima::DBI.  See that module's man page for
 details.
@@ -247,6 +252,16 @@ $obj is an instance of Class built out of a hash reference.
                         NumExplodingSheep   => 1
                       });
 
+If the primary column is not in %data, new() will assume it is to be
+generated.  If a sequence() has been specified for this Class, it will
+use that.  Otherwise, it will assume the primary key has an
+AUTO_INCREMENT constraint on it and attempt to use that.
+
+If the class has declared relationships with foreign classes via
+hasa(), it can pass an object to new() for the value of that key.
+Class::DBI will Do The Right Thing.
+
+
 =cut
 
 __PACKAGE__->set_sql('MakeNewObj', <<"SQL", 'Main');
@@ -260,6 +275,10 @@ __PACKAGE__->set_sql('LastInsertID', <<'', 'Main');
 SELECT LAST_INSERT_ID()
 
 
+__PACKAGE__->set_sql('Nextval', <<'', 'Main');
+SELECT NEXTVAL ('%s')
+
+
 sub new {
     my($proto, $data) = @_;
     my($class) = ref $proto || $proto;
@@ -278,6 +297,33 @@ sub new {
     # You -must- have a table defined.
     assert( $self->table ) if DEBUG;
 
+    # If a primary key wasn't given, use the sequence if we have one.
+    if( $self->sequence && !_safe_exists($data, $primary_col) ) {
+        my $sth = $self->sql_Nextval($self->sequence);        
+        $sth->execute;
+        $data->{$primary_col} = $sth->fetch;
+    }
+
+    # Look for values which can be objects.
+    my $hasa_cols = $class->__hasa_columns || {};
+    $class->normalize_hash($hasa_cols);
+
+    while( my($col, $want_class) = each %$hasa_cols) {
+        if( _safe_exists($data, $col) && ref $data->{$col} ) {
+            my $obj = $data->{$col};
+            unless( $obj->isa($want_class) ) {
+                require Carp;
+                Carp::croak(sprintf <<CARP, $obj->isa($want_class));
+$class expects an object of class $want_class for $col.  Got %s.
+CARP
+
+            }
+
+            $data->{$col} = $obj->id;
+        }
+    }
+        
+
     eval {
         # Enter a new row into the database containing our object's
         # information.
@@ -287,6 +333,7 @@ sub new {
                                        );
         $sth->execute(values %$data);
 
+        # If we still don't have a primary key, try AUTO_INCREMENT.
         unless( _safe_exists($data, $primary_col) ) {
             $sth = $self->sql_LastInsertID;
             $sth->execute;
@@ -595,6 +642,8 @@ sub commit {
   $obj->rollback;
 
 Removes any changes you've made to this object since the last commit.
+Currently this simply reloads the values from the database.  This can
+have concurrency issues.
 
 If you're using autocommit this method will throw an exception.
 
@@ -603,6 +652,7 @@ If you're using autocommit this method will throw an exception.
 #'#
 sub rollback {
     my($self) = shift;
+    my($class) = ref $self;
 
     # rollback() is useless if autocommit is on.
     if( $self->autocommit ) {
@@ -613,10 +663,34 @@ sub rollback {
     # Shortcut if there are no changes to rollback.
     return SUCCESS unless $self->is_changed;
 
-    # Stick the original values back into the object.
-    @{$self}{keys %{$self->{__Changed}}} = values %{$self->{__Changed}};
+    # Retrieve myself from the database again.
+    my $data;
+    eval {
+        my $sth = $self->sql_GetMe(join(', ', $self->is_changed),
+                                   $self->table,
+                                   $self->columns('Primary')
+                                  );
+        $sth->execute($self->id);
+        $data = $sth->fetchrow_hashref;
+        $sth->finish;
+    };
+    if ($@) {
+        $self->DBIwarn($self->id, 'GetMe');
+        return;
+    }
+    
+    unless( defined $data ) {
+        require Carp;
+        Carp::carp("rollback failed for ".$self->id." of class $class.");
+        return;
+    }
 
-    # Dump the changes.
+    # Make sure what we got from the database is what was changed.
+    assert( join('', sort keys %$data) eq
+            join('', sort $self->is_changed) ) if DEBUG;
+
+    # Throw away our changes.
+    @{$self}{keys %$data} = values %$data;
     $self->{__Changed}    = {};
 
     return SUCCESS;
@@ -627,9 +701,9 @@ sub DESTROY {
     
     if( my @changes = $self->is_changed ) {
         require Carp;
-        &Carp::carp( $self->id .' in class '. ref $self .
-                     ' destroyed without saving changes to '.
-                     join(', ', @changes) . ".\n"
+        &Carp::carp( $self->id .' in class '. ref($self) .
+                     ' destroyed without saving changes to column(s) '.
+                     join(', ', map { "'$_'" } @changes) . ".\n"
                    );
     }
 }
@@ -723,9 +797,11 @@ sub set {
 
     my $value = shift;
 
-    # Store the original value for rollback purposes.
-    $self->{__Changed}{$key} = $self->get($key) unless 
-      exists $self->{__Changed}{$key};
+    # Note the change for commit/rollback purposes.
+    # We increment instead of setting to 1 because it might be useful
+    # to someone to know how many times a value has changed between
+    # commits.
+    $self->{__Changed}{$key}++;
 
     $self->SUPER::set($key, $value);
 
@@ -791,7 +867,6 @@ __PACKAGE__->mk_classdata('__table');
 
 sub table {
     my($proto) = shift;
-
     my($class) = ref $proto || $proto;
 
     no strict 'refs';
@@ -810,6 +885,46 @@ sub table {
 
 =pod
 
+=item B<sequence>
+
+  Class->sequence($sequence_name);
+  $sequence_name = Class->sequence;
+  $sequence_name = $obj->sequence;
+
+An accessor to get/set the name of a sequence for the primary key.
+
+    Class->columns(Primary => 'id');
+    Class->sequence('class_id_seq');
+
+Class::DBI will use the sequence to generate primary keys when objects
+are created yet the primary key is not specified.
+
+B<NOTE>: Class::DBI also supports AUTO_INCREMENT and similar semantics.
+
+=cut
+
+__PACKAGE__->mk_classdata('__sequence');
+
+sub sequence {
+    my($proto) = shift;
+    my($class) = ref $proto || $proto;
+
+    no strict 'refs';
+    
+    if(@_) {
+        if( ref $proto ) {
+            require Carp;
+            Carp::carp('It is prefered to call sequence() as a class method '.
+                       '[Class->sequence($seq)] rather than an object method '.
+                       '[$obj->sequence($seq)] when setting the sequence.');
+        }
+    }
+    
+    $class->__sequence(@_);
+}
+
+=pod
+
 =item B<columns>
 
   @all_columns  = $obj->columns;
@@ -874,6 +989,11 @@ sub columns {
 
         $class->normalize(\@columns);
 
+        foreach my $col (@columns) {
+            $class->add_fields(PROTECTED, $col) unless 
+              $class->is_field($col);
+        }
+
         # Group all these columns together in their group and All.
         # XXX Should this add to the group or overwrite?
         $class_columns->{$group} = { map { ($_=>1) } @columns };
@@ -883,6 +1003,8 @@ sub columns {
         # Force columns() to be overriden if necessary.
         $class->__columns($class_columns);
 
+        $class->_flush_col2group;
+
         return SUCCESS;
     }
     else {
@@ -976,7 +1098,146 @@ sub _get_col2group {
 
     return $col2group;
 }
-    
+
+sub _flush_col2group {
+    my($proto) = shift;
+    my($class) = ref $proto || $proto;
+
+    my $col2group = $class->__Col2Group;
+    %$col2group = ();
+}
+
+
+=pod
+
+=back
+
+=head2 Table relationships, Object relationships
+
+Often you'll want one object to contain other objects in your
+database, in the same way one table references another with foreign
+keys.  For example, say we decided we wanted to store more information
+about directors of our films.  You might set up a table...
+
+    CREATE TABLE Directors (
+        Name            VARCHAR(80),
+        Birthday        INTEGER,
+        IsInsane        BOOLEAN
+    )
+
+And put a Class::DBI subclass around it.
+
+    package Film::Directors;
+    use base qw(Class::DBI);
+
+    Film::Directors->table('Directors');
+    Film::Directors->columns(All    => qw( Name Birthday IsInsane ));
+    Film::Directors->columns(Prmary => qw( Name ));
+    Film::Directors->set_db(Main => 'dbi:mysql', 'me', 'heywoodjablowme',
+                            {AutoCommit => 1});
+
+Now Film can use its Director column as a way of getting at
+Film::Directors objects, instead of just the director's name.  Its a
+simple matter of adding one line to Film.
+
+    # Director() is now an accessor to Film::Directors objects.
+    Film->hasa('Film::Directors', 'Director');
+
+Now the Film->Director() accessor gets and sets Film::Director objects
+instead of just their name.
+
+=over 4
+
+=item B<hasa>
+
+    Class->hasa($foreign_class, @foreign_key_columns);
+
+Declares that the given Class has a relationship with the
+$foreign_class and is storing $foreign_class's primary key
+information in the @foreign_key_columns.
+
+An accessor will be generated with the name of the first element in
+@foreign_key_columns.  It gets/sets objects of $foreign_class.  Using
+our Film::Director example...
+
+    # Set the director of Bad Taste to the Film::Director object
+    # representing Peter Jackson.
+    $pj     = Film::Directory->retreive('Peter Jackson');
+    $btaste = Film->retreive('Bad Taste');
+    $btaste->Director($pj);
+
+hasa() will try to require the foreign class for you.  If the require
+fails, it will assume its not a simple require (ie. Foreign::Class
+isn't in Foreign/Class.pm) and that you've already taken care of it
+and ignore the warning.
+
+It is not necessary to call columns() to set up the
+@foreign_key_columns.  hasa() will do this for you if you haven't
+already.
+
+XXX I don't know if I like the way this works.  It may change a bit in
+the future.  I'm not sure about the way the accessor is named.
+
+NOTE  The two classes do not have to be in the same database!
+
+=cut
+
+#'#
+__PACKAGE__->mk_classdata('__hasa_columns');
+
+sub hasa {
+    my($class, $foreign_class, @foreign_key_cols) = @_;
+
+    my $foreign_col_accessor = "_".$foreign_key_cols[0]."_accessor";
+
+    eval "require $foreign_class";
+
+    # This is so complicated to allow multiple columns leading to the
+    # same class.
+    my $obj_key = "__".$foreign_class."_". 
+                  join(':', @foreign_key_cols)."_Obj";
+
+    # Setup the columns for this foreign class.
+    $class->columns($obj_key, @foreign_key_cols);
+
+    # Make sure pseudohashes know about the object key field.
+    $class->add_fields(PROTECTED, $obj_key);
+
+    my $hasa_columns = $class->__hasa_columns || {};
+    @{$hasa_columns}{@foreign_key_cols} = 
+        ($foreign_class) x @foreign_key_cols;
+
+    $class->__hasa_columns($hasa_columns);
+
+    my $accessor = sub {
+        my($self) = shift;
+        
+        if ( @_ ) {             # setting
+            my($obj) = shift;
+            $self->{$obj_key} = $obj;
+            
+            # XXX Have to fix this for mult-col foreign keys.
+            $self->$foreign_col_accessor($obj->id);
+        }
+        
+        unless ( defined $self->{$obj_key} ) {
+            # XXX Fix this, too.
+            my $obj_id = $self->$foreign_col_accessor();
+            $self->{$obj_key} = $foreign_class->retrieve($obj_id) if
+              defined $obj_id;
+        }
+        
+        return $self->{$obj_key};
+    };
+      
+    # This might cause a subroutine redefined warning.
+    {
+        local $^W = 0;
+        no strict 'refs';
+        *{$class."\::$foreign_key_cols[0]"} = $accessor;
+    }
+}
+
 
 =pod
 
@@ -1035,7 +1296,7 @@ as data keys.
 sub normalize {
     my($self, $columns) = @_;
     
-    assert(@$columns > 0) if DEBUG;
+    assert(ref $columns eq 'ARRAY') if DEBUG;
 
     foreach my $col (@$columns) {
         $col =~ s/^.*\.//;  # Chop off the possible table & database names.
@@ -1264,6 +1525,9 @@ Having more than one column as your primary key in the SQL table is
 currently not supported.  Why?  Its more complicated.  A later version
 will support multi-column keys.
 
+
+=head1 TODO
+
 =head2 Table/object relationships need to be handled.
 
 There's no graceful way to handle relationships between two
@@ -1275,17 +1539,30 @@ fairly simple manner.
 There's no graceful way to handle lists of things as object data.
 This is also something I plan to implement eventually.
 
+=head2 Using pseudohashes as objects has to be documented
+
+=head2 Cookbook needs to be written
+
+=head2 Object caching needs to be added
+
+=head2 Multi-column primary keys untested.
+
+If you need this feature let me know and I'll get it working.
+
+=head2 More testing with more databases.
+
+=head2 Complex data storage via Storable needed.
+
+=head2 There are concurrency problems
+
+=head2 rollback() has concurrency problems
+
 
 =head1 BUGS
 
-=head2 Untested in production.
-
-I've yet to really throw this into a production environment and shake
-out the bugs.  I will be using it in that capacity soon and it will
-firm up quickly.
-
-=head2 Only tested with DBD::CSV and MySQL
+=head2 Only tested with DBD::CSV and DBD::mysql.
 
+=head2 sequence() has not been tested.
 
 =head1 AUTHOR
 
@@ -1296,5 +1573,7 @@ Uri Gutman and Damian Conway.
 =head1 SEE ALSO
 
 L<Ima::DBI>, L<Class::Accessor>, L<base>, L<Class::Data::Inheritable>
+http://www.pobox.com/~schwern/papers/Class-DBI/,
+Perl Object-Oriented Persistence E<lt>poop-group@lists.sourceforge.netE<gt>
 
 =cut
